import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

from unittest.mock import Mock, call

import pytest

import jj
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher
from jj.middlewares import BaseMiddleware, LoggerMiddleware
from jj.resolvers import Registry, ReversedResolver
from jj.responses import Response

from .._test_utils import run


class TestLoggerMiddleware(TestCase):
    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    def test_middleware_without_resolver(self):
        with self.assertRaises(Exception):
            LoggerMiddleware()

    def test_middleware_without_logger(self):
        with self.assertRaises(Exception):
            LoggerMiddleware(self.resolver)

    def test_middleware_with_instance_resolver(self):
        logger = Mock()
        middleware = LoggerMiddleware(self.resolver, logger)
        self.assertIsInstance(middleware, BaseMiddleware)

    @pytest.mark.asyncio
    async def test_app_logger(self):
        mock = Mock()
        record = {"request": None, "response": None}

        @LoggerMiddleware(self.resolver, mock)
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                record["request"] = request
                response = Response(status=200)
                record["response"] = response
                return response

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_has_calls([
            call.info(record["request"], extra={
                "jj_request": record["request"],
            }),
            call.info(record["response"], extra={
                "jj_request": record["request"],
                "jj_response": record["response"],
            }),
        ])
        self.assertEqual(mock.info.call_count, 2)

    @pytest.mark.asyncio
    async def test_handler_logger(self):
        mock = Mock()
        record = {"request": None, "response": None}

        class App(jj.App):
            resolver = self.resolver

            @LoggerMiddleware(self.resolver, mock)
            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                record["request"] = request
                response = Response(status=200)
                record["response"] = response
                return response

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_has_calls([
            call.info(record["request"], extra={
                "jj_request": record["request"],
            }),
            call.info(record["response"], extra={
                "jj_request": record["request"],
                "jj_response": record["response"],
            }),
        ])
        self.assertEqual(mock.info.call_count, 2)

    @pytest.mark.asyncio
    async def test_app_and_handler_logger(self):
        app_logger, handler_logger = Mock(), Mock()

        @LoggerMiddleware(self.resolver, app_logger)
        class App(jj.App):
            resolver = self.resolver

            @LoggerMiddleware(resolver, handler_logger)
            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(status=200)

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        self.assertEqual(app_logger.info.call_count, 0)
        self.assertEqual(handler_logger.info.call_count, 2)

    @pytest.mark.asyncio
    async def test_handler_without_logger(self):
        class App(jj.App):
            resolver = self.resolver

            @LoggerMiddleware(self.resolver, None)
            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                response = Response(status=200)
                return response

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

import pytest

import jj
from jj.apps import DefaultApp, create_app
from jj.handlers import default_handler
from jj.matchers import PathMatcher
from jj.resolvers import Registry, ReversedResolver
from jj.responses import Response

from .._test_utils import run


class TestDefaultApp(TestCase):
    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    def test_default_app_is_singleton(self):
        self.assertEqual(DefaultApp(), DefaultApp())

    @pytest.mark.asyncio
    async def test_default_app_with_handler(self):
        path, status, text = "/route", 201, "text"

        @PathMatcher(path, resolver=self.resolver)
        async def handler(request):
            return Response(status=status, text=text)

        async with run(self.default_app, self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status)
            self.assertEqual(await response.text(), text)

    @pytest.mark.asyncio
    async def test_default_app_without_handlers(self):
        path, status, text = "/route", 201, "text"

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path, resolver=resolver)
            async def handler(request):
                return Response(status=status, text=text)

        async with run(self.default_app, self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, 404)

    @pytest.mark.asyncio
    async def test_default_app(self):
        path = "/route"
        status1, text1 = 201, "text-1"
        status2, text2 = 202, "text-2"

        @PathMatcher(path, resolver=self.resolver)
        async def handler(request):
            return Response(status=status1, text=text1)

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path, resolver=resolver)
            async def handler(request):
                return Response(status=status2, text=text2)

        async with run(self.default_app, self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status1)
            self.assertEqual(await response.text(), text1)

        async with run(App(), self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status2)
            self.assertEqual(await response.text(), text2)



from unittest.mock import sentinel

import pytest
from pytest import raises

from jj.matchers import AttributeMatcher

from ..._test_utils.steps import given, then, when


@pytest.mark.asyncio
async def test_abstract_match_method_raises_exception():
    with given:
        matcher = AttributeMatcher()

    with when, raises(Exception) as exception:
        await matcher.match(sentinel.value)

    with then:
        assert exception.type is NotImplementedError


def test_repr():
    with given:
        matcher = AttributeMatcher()

    with when:
        actual = repr(matcher)

    with then:
        assert actual == "AttributeMatcher()"



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call

import pytest
from pytest import raises

from jj.matchers import ResolvableMatcher

from .._test_utils.fixtures import handler_, request_, resolver_
from .._test_utils.steps import given, then, when

__all__ = ("request_", "resolver_", "handler_",)


@pytest.mark.asyncio
async def test_abstract_match_method_raises_error(*, resolver_, request_):
    with given:
        matcher = ResolvableMatcher(resolver=resolver_)

    with when, raises(Exception) as exception:
        await matcher.match(request_)

    with then:
        assert exception.type is NotImplementedError


@pytest.mark.asyncio
async def test_concrete_match_method_not_raises_error(*, resolver_, request_):
    with given:
        matcher_ = Mock(ResolvableMatcher, match=AsyncMock(return_value=True))

    with when:
        actual = await matcher_.match(request_)

    with then:
        assert actual is True
        assert matcher_.mock_calls == [call.match(request_)]


@pytest.mark.asyncio
async def test_decorator_registers_matcher(*, resolver_, handler_):
    with given:
        matcher = ResolvableMatcher(resolver=resolver_)

    with when:
        actual = matcher(handler_)

    with then:
        assert actual == handler_
        assert resolver_.mock_calls == [call.register_matcher(matcher.match, handler_)]
        handler_.assert_not_called()


def test_repr(*, resolver_):
    with given:
        resolver_.__repr__ = Mock(return_value="<Resolver>")
        matcher = ResolvableMatcher(resolver=resolver_)

    with when:
        actual = repr(matcher)

    with then:
        assert actual == "ResolvableMatcher(resolver=<Resolver>)"



# 1xx

CONTINUE = 100
SWITCHING_PROTOCOLS = 101

# 2xx
OK = 200
CREATED = 201
ACCEPTED = 202
NON_AUTHORITATIVE_INFORMATION = 203
NO_CONTENT = 204
RESET_CONTENT = 205
PARTIAL_CONTENT = 206

# 3xx
MULTIPLE_CHOICES = 300
MOVED_PERMANENTLY = 301
FOUND = 302
SEE_OTHER = 303
NOT_MODIFIED = 304
TEMPORARY_REDIRECT = 307
PERMANENT_REDIRECT = 308

# 4xx
BAD_REQUEST = 400
UNAUTHORIZED = 401
FORBIDDEN = 403
NOT_FOUND = 404
METHOD_NOT_ALLOWED = 405
NOT_ACCEPTABLE = 406
PROXY_AUTHENTICATION_REQUIRED = 407
REQUEST_TIMEOUT = 408
CONFLICT = 409
GONE = 410
LENGTH_REQUIRED = 411
PRECONDITION_FAILED = 412
PAYLOAD_TOO_LARGE = 413
URI_TOO_LONG = 414
UNSUPPORTED_MEDIA_TYPE = 415
RANGE_NOT_SATISFIABLE = 416
EXPECTATION_FAILED = 417
I_AM_A_TEAPOT = 418
UNPROCESSABLE_ENTITY = 422
TOO_EARLY = 425
UPGRADE_REQUIRED = 426
PRECONDITION_REQUIRED = 428
TOO_MANY_REQUESTS = 429
REQUEST_HEADER_FIELDS_TOO_LARGE = 431
UNAVAILABLE_FOR_LEGAL_REASONS = 451

# 5xx
INTERNAL_SERVER_ERROR = 500
NOT_IMPLEMENTED = 501
BAD_GATEWAY = 502
SERVICE_UNAVAILABLE = 503
GATEWAY_TIMEOUT = 504
HTTP_VERSION_NOT_SUPPORTED = 505
NETWORK_AUTHENTICATION_REQUIRED = 511



from unittest.mock import sentinel

from jj.matchers import AttributeMatcher
from jj.matchers.attribute_matchers import RouteMatcher

from ..._test_utils.steps import given, then, when


def test_is_instance_of_attribute_matcher():
    with given:
        matcher = RouteMatcher("/")

    with when:
        actual = isinstance(matcher, AttributeMatcher)

    with then:
        assert actual is True


def test_repr():
    with given:
        matcher = RouteMatcher("/")

    with when:
        actual = repr(matcher)

    with then:
        assert actual == "RouteMatcher('/')"


def test_pack():
    with given:
        path = "/"
        matcher = RouteMatcher(path)

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {"path": path}


def test_unpack():
    with given:
        kwargs = {
            "path": "/",
            "future_field": sentinel,
        }

    with when:
        actual = RouteMatcher.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, RouteMatcher)



from unittest.mock import Mock

import pytest
from pytest import raises

from jj.matchers import LogicalMatcher, ResolvableMatcher

from ..._test_utils.fixtures import request_, resolver_
from ..._test_utils.steps import given, then, when

__all__ = ("request_", "resolver_",)


@pytest.mark.asyncio
async def test_abstract_match_method_raises_exception(*, resolver_, request_):
    with given:
        matcher = LogicalMatcher(resolver=resolver_)

    with when, raises(Exception) as exception:
        await matcher.match(request_)

    with then:
        assert exception.type is NotImplementedError


def test_is_instance_of_resolvable_matcher(*, resolver_):
    with given:
        matcher = LogicalMatcher(resolver=resolver_)

    with when:
        actual = isinstance(matcher, ResolvableMatcher)

    with then:
        assert actual is True


def test_repr(*, resolver_):
    with given:
        resolver_.__repr__ = Mock(return_value="<Resolver>")
        matcher = LogicalMatcher(resolver=resolver_)

    with when:
        actual = repr(matcher)

    with then:
        assert actual == "LogicalMatcher(resolver=<Resolver>)"



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

from unittest.mock import Mock, call, sentinel

import pytest

import jj
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher
from jj.middlewares import AbstractMiddleware, RootMiddleware
from jj.requests import Request
from jj.resolvers import Registry, ReversedResolver
from jj.responses import Response

from .._test_utils import run


class TestRootMiddleware(TestCase):
    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    def test_middleware_without_resolver(self):
        with self.assertRaises(Exception):
            RootMiddleware()

    def test_middleware_with_instance_resolver(self):
        middleware = RootMiddleware(self.resolver)
        self.assertIsInstance(middleware, AbstractMiddleware)

    def test_middleware_with_class_resolver(self):
        class Middleware(RootMiddleware):
            resolver = self.resolver
        middleware = Middleware()
        self.assertIsInstance(middleware, AbstractMiddleware)

    @pytest.mark.asyncio
    async def test_middleware_without_impl(self):
        class App(jj.App):
            resolver = self.resolver

        middleware = RootMiddleware(self.resolver)
        async with run(App(), middlewares=[middleware]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 404)

    @pytest.mark.asyncio
    async def test_middleware_without_impl_but_with_handlers(self):
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(status=200)

        middleware = RootMiddleware(self.resolver)
        async with run(App(), middlewares=[middleware]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

    @pytest.mark.asyncio
    async def test_middleware_with_impl(self):
        mock = Mock()

        class App(jj.App):
            resolver = self.resolver

        class Middleware(RootMiddleware):
            async def do(self, r, h, a):
                mock(r, h, a)
                return await super().do(r, h, a)

        app, middleware = App(), Middleware(self.resolver)
        async with run(app, middlewares=[middleware]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 404)

        mock.assert_called_once()
        request_arg, handler_arg, app_arg = mock.call_args[0]
        self.assertIsInstance(request_arg, Request)
        self.assertEqual(handler_arg, default_handler)
        self.assertEqual(app_arg, app)

    @pytest.mark.asyncio
    async def test_multiple_root_middlewares(self):
        mock = Mock()

        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                mock(App.__name__, sentinel.BEFORE)
                response = Response(status=200)
                mock(App.__name__, sentinel.AFTER)
                return response

        class Middleware1(RootMiddleware):
            async def do(self, request, handler, app):
                mock(self.__class__.__name__, sentinel.BEFORE)
                response = await super().do(request, handler, app)
                mock(self.__class__.__name__, sentinel.AFTER)
                return response

        class Middleware2(Middleware1):
            pass

        middleware1, middleware2 = Middleware1(self.resolver), Middleware2(self.resolver)
        async with run(App(), middlewares=[middleware2, middleware1]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_has_calls([
            call(Middleware2.__name__, sentinel.BEFORE),
            call(Middleware1.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.AFTER),
            call(Middleware1.__name__, sentinel.AFTER),
            call(Middleware2.__name__, sentinel.AFTER),
        ])
        self.assertEqual(mock.call_count, 6)

    @pytest.mark.asyncio
    async def test_same_root_middlewares(self):
        mock = Mock()

        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(status=200)

        class Middleware(RootMiddleware):
            async def do(self, request, handler, app):
                mock()
                return await super().do(request, handler, app)

        middleware = Middleware(self.resolver)
        async with run(App(), middlewares=[middleware, middleware]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_called_once_with()



import pytest

import jj
from jj.middlewares import SelfMiddleware
from jj.mock import Mock, Mocked, RemoteMock

from .._test_utils import run


@pytest.mark.asyncio
async def test_mocked_prefetch_history():
    remote_mock = Mock()
    self_middleware = SelfMiddleware(remote_mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200)

    async with run(remote_mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)

        async with Mocked(handler, prefetch_history=True) as mock:
            response = await client.get("/")
            assert response.status == 200
            assert mock.history is None

        assert len(mock.history) == 1


@pytest.mark.asyncio
async def test_mocked_not_prefetch_history():
    remote_mock = Mock()
    self_middleware = SelfMiddleware(remote_mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200)

    async with run(remote_mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)

        async with Mocked(handler, prefetch_history=False) as mock:
            response = await client.get("/")
            assert response.status == 200

        assert mock.history is None


@pytest.mark.asyncio
async def test_mocked_fetch_history():
    remote_mock = Mock()
    self_middleware = SelfMiddleware(remote_mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200)

    async with run(remote_mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)

        async with Mocked(handler, prefetch_history=False) as mock:
            response = await client.get("/")
            assert response.status == 200
            assert mock.history is None

            history = await mock.fetch_history()
            assert len(history) == len(mock.history) == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("prefetch_history", [True, False])
async def test_mocked_wait_for_requests(*, prefetch_history: bool):
    remote_mock = Mock()
    self_middleware = SelfMiddleware(remote_mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200)

    async with run(remote_mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)

        async with Mocked(handler, prefetch_history=prefetch_history) as mock:
            response = await client.get("/")
            assert response.status == 200
            await mock.wait_for_requests()

        assert len(mock.history) == 1


@pytest.mark.asyncio
async def test_mocked_wait_for_requests_timeout():
    remote_mock = Mock()
    self_middleware = SelfMiddleware(remote_mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200)

    async with run(remote_mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)

        async with Mocked(handler) as mock:
            await mock.wait_for_requests(timeout=0.01)
            assert len(mock.history) == 0


@pytest.mark.asyncio
@pytest.mark.parametrize(("disposable", "status"), [
    (True, 404),
    (False, 200),
])
async def test_mocked_disposable(*, disposable: bool, status: int):
    remote_mock = Mock()
    self_middleware = SelfMiddleware(remote_mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200)

    async with run(remote_mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)

        async with Mocked(handler, disposable=disposable):
            response = await client.get("/")
            assert response.status == 200

        response = await client.get("/")
        assert response.status == status


@pytest.mark.asyncio
@pytest.mark.parametrize(("disposable", "prefetch_history"), [
    (True, False),
    (False, True),
])
async def test_mocked_repr(*, disposable: bool, prefetch_history: bool):
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)
    matcher, response = jj.match("*"), jj.Response()

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        mocked = Mocked(handler, disposable=disposable, prefetch_history=prefetch_history)
        assert repr(mocked) == (f"Mocked<{handler}, "
                                f"disposable={disposable}, "
                                f"prefetch_history={prefetch_history}>")



from ._expiration_policy import ExpirationPolicy
from ._expire_after_requests import ExpireAfterRequests
from ._expire_never import ExpireNever

__all__ = (
    "ExpirationPolicy",
    "ExpireNever",
    "ExpireAfterRequests",
)



from typing import Callable

from ._body_parser import BodyParser
from ._history_item import HistoryItem

__all__ = ("default_history_adapter", "HistoryAdapterType")

HistoryAdapterType = Callable[[HistoryItem], HistoryItem]

default_history_adapter = BodyParser().parse



from jj.expiration_policy import ExpireAfterRequests, ExpireNever

from .._test_utils.steps import given, then, when


def test_unpack_expire_never():
    with given:
        packed = {}

    with when:
        expiration_policy = ExpireNever.__unpacked__(**packed)

    with then:
        assert expiration_policy.__packed__() == packed


def test_unpack_expire_after_requests():
    with given:
        packed = {
            "max_requests_count": 2
        }

    with when:
        expiration_policy = ExpireAfterRequests.__unpacked__(**packed)

    with then:
        assert expiration_policy.__packed__() == packed



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

import pytest

import jj
from jj import server_version
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher
from jj.resolvers import Registry, ReversedResolver
from jj.responses import StreamResponse

from .._test_utils import run


class TestStreamResponse(TestCase):
    def make_app_with_response(self, *args, **kwargs):
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return StreamResponse(*args, **kwargs)
        return App()

    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    @pytest.mark.asyncio
    async def test_response_with_manual_preparing(self):
        body = b"200 OK"

        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                stream = StreamResponse()
                await stream.prepare(request)
                await stream.write(body)
                await stream.write_eof()
                return stream

        async with run(App()) as client:
            response = await client.get("/")
            # status
            self.assertEqual(response.status, 200)
            self.assertEqual(response.reason, "OK")
            # headers
            self.assertEqual(response.headers.get("Server"), server_version)
            self.assertEqual(response.headers.get("Content-Type"), "application/octet-stream")
            self.assertEqual(response.headers.get("Transfer-Encoding"), "chunked")
            self.assertIsNotNone(response.headers.get("Date"))
            self.assertEqual(len(response.headers), 4)
            # body
            raw = await response.read()
            self.assertEqual(raw, body)

    @pytest.mark.asyncio
    async def test_response_with_default_args(self):
        app = self.make_app_with_response()

        async with run(app) as client:
            response = await client.get("/")
            # status
            self.assertEqual(response.status, 200)
            self.assertEqual(response.reason, "OK")
            # headers
            self.assertEqual(response.headers.get("Server"), server_version)
            self.assertEqual(response.headers.get("Content-Type"), "application/octet-stream")
            self.assertEqual(response.headers.get("Transfer-Encoding"), "chunked")
            self.assertIsNotNone(response.headers.get("Date"))
            self.assertEqual(len(response.headers), 4)
            # body
            raw = await response.read()
            self.assertEqual(raw, b"")

    # Status

    @pytest.mark.asyncio
    async def test_response_status(self):
        status = 204
        app = self.make_app_with_response(status=status)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.status, status)
            # aiohttp автоматически подставляет нужный reason
            self.assertEqual(response.reason, "No Content")

    @pytest.mark.asyncio
    async def test_response_reason(self):
        status, reason = 204, "Custom Reason"
        app = self.make_app_with_response(status=status, reason=reason)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.status, status)
            self.assertEqual(response.reason, reason)

    # Headers

    @pytest.mark.asyncio
    async def test_response_headers(self):
        custom_header_key, custom_header_val = "Cutom-Header", "Value"
        app = self.make_app_with_response(headers={custom_header_key: custom_header_val})

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get(custom_header_key), custom_header_val)
            self.assertEqual(response.headers.get("Server"), server_version)

    @pytest.mark.asyncio
    async def test_response_with_custom_server_header(self):
        server_header = "server version x"
        app = self.make_app_with_response(headers={"Server": server_header})

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get("Server"), server_header)

    @pytest.mark.asyncio
    async def test_response_with_expect_header(self):
        app = self.make_app_with_response()

        async with run(app) as client:
            response = await client.post("/", json={}, expect100=True)
            self.assertEqual(response.status, 200)

    @pytest.mark.asyncio
    async def test_response_with_incorrect_expect_header(self):
        app = self.make_app_with_response()

        async with run(app) as client:
            response = await client.post("/", json={}, headers={"Expect": "banana"})
            self.assertEqual(response.status, 417)



import sys

__all__ = ("HistoryItem",)


if sys.version_info >= (3, 8):
    from typing import List, TypedDict

    from ._history_request import HistoryRequest
    from ._history_response import HistoryResponse

    class HistoryItem(TypedDict):
        request: HistoryRequest
        response: HistoryResponse
        tags: List[str]
else:
    from typing import Any, Dict
    HistoryItem = Dict[str, Any]



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call, sentinel

import pytest
from pytest import raises

from jj.matchers import AllMatcher, LogicalMatcher, ResolvableMatcher

from ..._test_utils.fixtures import request_, resolver_
from ..._test_utils.steps import given, then, when

__all__ = ("request_", "resolver_",)


@pytest.mark.asyncio
@pytest.mark.parametrize(("ret_val", "res"), [
    (True, True),
    (False, False),
])
async def test_single_submatcher(ret_val, res, *, resolver_, request_):
    with given:
        submatcher_ = Mock(ResolvableMatcher, match=AsyncMock(return_value=ret_val))
        matcher = AllMatcher([submatcher_], resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is res
        assert submatcher_.mock_calls == [call.match(request_)]


@pytest.mark.asyncio
@pytest.mark.parametrize(("ret_val1", "ret_val2", "res"), [
    (True, True, True),
    (True, False, False),
])
async def test_multiple_truthy_submatchers(ret_val1, ret_val2, res, *, resolver_, request_):
    with given:
        submatcher1_ = Mock(ResolvableMatcher, match=AsyncMock(return_value=ret_val1))
        submatcher2_ = Mock(ResolvableMatcher, match=AsyncMock(return_value=ret_val2))
        matcher = AllMatcher([submatcher1_, submatcher2_], resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is res
        assert submatcher1_.mock_calls == [call.match(request_)]
        assert submatcher2_.mock_calls == [call.match(request_)]


@pytest.mark.asyncio
@pytest.mark.parametrize(("ret_val1", "ret_val2", "res"), [
    (False, True, False),
    (False, False, False),
])
async def test_multiple_false_submatchers(ret_val1, ret_val2, res, *, resolver_, request_):
    with given:
        submatcher1_ = Mock(ResolvableMatcher, match=AsyncMock(return_value=ret_val1))
        submatcher2_ = Mock(ResolvableMatcher, match=AsyncMock(return_value=ret_val2))
        matcher = AllMatcher([submatcher1_, submatcher2_], resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is res
        assert submatcher1_.mock_calls == [call.match(request_)]
        assert submatcher2_.mock_calls == []


def test_empty_submatchers_raises_exception(*, resolver_):
    with when, raises(Exception) as exception:
        AllMatcher([], resolver=resolver_)

    with then:
        assert exception.type is AssertionError


def test_is_instance_of_logical_matcher(*, resolver_):
    with given:
        submatcher_ = Mock(ResolvableMatcher)
        matcher = AllMatcher([submatcher_], resolver=resolver_)

    with when:
        actual = isinstance(matcher, LogicalMatcher)

    with then:
        assert actual is True


def test_repr(*, resolver_):
    with given:
        resolver_.__repr__ = Mock(return_value="<Resolver>")
        matcher = AllMatcher(resolver=resolver_, matchers=[
            Mock(ResolvableMatcher, __repr__=Mock(return_value="<SubMatcher1>")),
            Mock(ResolvableMatcher, __repr__=Mock(return_value="<SubMatcher2>")),
        ])

    with when:
        actual = repr(matcher)

    with then:
        assert actual == "AllMatcher([<SubMatcher1>, <SubMatcher2>], resolver=<Resolver>)"


def test_pack(*, resolver_):
    with given:
        submatchers = [Mock(ResolvableMatcher), Mock(ResolvableMatcher)]
        matcher = AllMatcher(submatchers, resolver=resolver_)

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {"matchers": submatchers}


def test_unpack(*, resolver_):
    with given:
        submatchers = [Mock(ResolvableMatcher), Mock(ResolvableMatcher)]
        kwargs = {
            "matchers": submatchers,
            "resolver": resolver_,
            "future_field": sentinel,
        }

    with when:
        actual = AllMatcher.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, AllMatcher)



from typing import Type, Union

from ..apps import AbstractApp
from ..handlers import HandlerFunction
from ._abstract_middleware import AbstractMiddleware
from ._middleware_type import MiddlewareType

__all__ = ("BaseMiddleware", "AppOrHandler",)


AppOrHandler = Union[Type[AbstractApp], HandlerFunction]


class BaseMiddleware(AbstractMiddleware):
    def on_app(self, app: Type[AbstractApp]) -> None:
        pass

    def on_handler(self, handler: HandlerFunction) -> None:
        pass

    def _call_hooks(self, app_or_handler: AppOrHandler) -> None:
        if isinstance(app_or_handler, type) and issubclass(app_or_handler, AbstractApp):
            return self.on_app(app_or_handler)
        return self.on_handler(app_or_handler)

    def _register_middleware(self,
                             app_or_handler: AppOrHandler, middleware: MiddlewareType) -> None:
        old_middlewares = self._resolver.get_attribute("middlewares", app_or_handler, [])
        new_middlewares = old_middlewares + [middleware]
        self._resolver.register_attribute("middlewares", new_middlewares, app_or_handler)

    def __call__(self, app_or_handler: AppOrHandler) -> AppOrHandler:
        self._call_hooks(app_or_handler)
        self._register_middleware(app_or_handler, self._do)
        return app_or_handler



import logging

import pytest

from jj.logs import Logger

from .._test_utils.steps import given, then, when


@pytest.fixture()
def logger():
    return Logger(__name__)


def test_no_default_handlers(logger: Logger):
    with when:
        assert len(logger.handlers) == 0


def test_add_handler(logger: Logger):
    with given:
        handler = logging.NullHandler()

    with when:
        logger.addHandler(handler)

    with then:
        assert len(logger.handlers) == 1


def test_clear_handlers(logger: Logger):
    with given:
        handler = logging.NullHandler()
        logger.addHandler(handler)

    with when:
        res = logger.clearHandlers()

    with then:
        assert res == logger
        assert len(logger.handlers) == 0



from io import BufferedReader, BytesIO, StringIO, TextIOWrapper

import pytest
from aiohttp.web import ContentCoding
from pytest import raises

from jj._version import server_version
from jj.responses import Response

from .._test_utils.steps import given, then, when


def default_packed(headers, status=200, reason="OK", body=b"",
                   cookies=None, chunked=False, compression=None):
    return {
        "status": status,
        "reason": reason,
        "body": body,
        "cookies": cookies or [],
        "headers": headers,
        "chunked": chunked,
        "compression": compression,
    }


def test_pack_text_body():
    with given:
        text = "200 OK"
        binary_text = b"200 OK"
        response = Response(text=text)

    with when:
        actual = response.__packed__()

    with then:
        assert actual == default_packed(body=binary_text, headers=[
            ["Content-Type", "text/plain; charset=utf-8"],
            ["Server", server_version],
        ])


def test_pack_json_body():
    with given:
        json = {}
        binary_json = b"{}"
        response = Response(json=json)

    with when:
        actual = response.__packed__()

    with then:
        assert actual == default_packed(body=binary_json, headers=[
            ["Content-Type", "application/json"],
            ["Server", server_version],
            ["Content-Length", str(len(binary_json))],
        ])


@pytest.mark.parametrize(("body", "expected", "headers"), [
    (None, b"", [
        ["Server", server_version],
        ["Content-Length", "0"],
        ["Content-Type", "text/plain; charset=utf-8"],
    ]),
    (b"text", b"text", [
        ["Server", server_version],
    ]),
    (bytearray(b"text"), b"text", [
        ["Server", server_version],
    ]),
    (memoryview(b"text"), b"text", [
        ["Server", server_version],
        ["Content-Length", "4"],
        ["Content-Type", "application/octet-stream"],
    ]),
    ("text", b"text", [
        ["Server", server_version],
        ["Content-Length", "4"],
        ["Content-Type", "text/plain; charset=utf-8"],
    ]),
    (StringIO("text"), b"text", [
        ["Content-Disposition", "inline"],
        ["Server", server_version],
        ["Content-Length", "4"],
        ["Content-Type", "text/plain; charset=utf-8"],
    ]),
    (BytesIO(b"text"), b"text", [
        ["Content-Disposition", "inline"],
        ["Server", server_version],
        ["Content-Length", "4"],
        ["Content-Type", "application/octet-stream"],
    ]),
    (TextIOWrapper(BytesIO(b"text")), b"text", [
        ["Content-Disposition", "inline"],
        ["Server", server_version],
        ["Content-Type", "text/plain; charset=utf-8"],
    ]),
    (BufferedReader(BytesIO(b"text")), b"text", [
        ["Content-Disposition", "inline"],
        ["Server", server_version],
        ["Content-Type", "application/octet-stream"],
    ]),
])
def test_pack_body(body, expected, headers):
    with given:
        response = Response(body=body)

    with when:
        actual = response.__packed__()

    with then:
        assert actual == default_packed(body=expected, headers=headers)


def test_pack_unsupported_body():
    with given:
        class AsyncIterable:
            def __aiter__(self):
                return self

            async def __anext__(self):
                raise StopAsyncIteration

        response = Response(body=AsyncIterable())

    with when, raises(Exception) as exception:
        response.__packed__()

    with then:
        assert exception.type is ValueError


def test_pack_chunked():
    with given:
        body = b"200 OK"
        response = Response(body=body)
        response.enable_chunked_encoding()

    with when:
        actual = response.__packed__()

    with then:
        assert actual == default_packed(body=body, chunked=True, headers=[
            ["Server", server_version],
        ])


def test_pack_compressed():
    with given:
        body = b"200 OK"
        response = Response(body=body)
        response.enable_compression(ContentCoding.gzip)

    with when:
        actual = response.__packed__()

    with then:
        assert actual == default_packed(body=body, compression=ContentCoding.gzip.value, headers=[
            ["Server", server_version],
        ])


def test_pack_cookies():
    with given:
        body = b"200 OK"
        response = Response(body=body)
        cookie = {
            "name": "name1",
            "value": "value1",
            "expires": "Thu, 01 Jan 1970 00:00:01 GMT",
            "domain": ".localhost",
            "max_age": "0",
            "path": "/path",
            "secure": "secure",
            "httponly": "httponly",
            "version": "1",
        }
        response.set_cookie(**cookie)
        response.set_cookie("name2", "value2")

    with when:
        actual = response.__packed__()

    with then:
        assert actual == default_packed(body=body, headers=[["Server", server_version]], cookies=[
            cookie,
            {
                "name": "name2",
                "value": "value2",
                "expires": None,
                "domain": None,
                "max_age": None,
                "path": "/",
                "secure": None,
                "httponly": None,
                "version": None,
            }
        ])


def test_pack_text_body_with_headers():
    with given:
        text = "200 OK"
        binary_text = b"200 OK"
        response = Response(text=text, headers=[
            ("X-Custom-Header1", "Value1"),
            ("X-Custom-Header2", "Value2"),
        ])

    with when:
        actual = response.__packed__()

    with then:
        assert actual == default_packed(body=binary_text, headers=[
            ["X-Custom-Header1", "Value1"],
            ["X-Custom-Header2", "Value2"],
            ["Content-Type", "text/plain; charset=utf-8"],
            ["Server", server_version],
        ])



import pytest
from packed import pack

from jj.matchers import (
    ContainMatcher,
    EqualMatcher,
    ExistMatcher,
    NotContainMatcher,
    NotEqualMatcher,
    RegexMatcher,
    contains,
    equals,
    exists,
    not_contains,
    not_equals,
    regex,
)

from ..._test_utils.steps import given, then, when


@pytest.mark.parametrize(("instance_factory", "instance_class"), [
    (lambda: exists, ExistMatcher),
    (lambda: equals("smth"), EqualMatcher),
    (lambda: not_equals("smth"), NotEqualMatcher),
    (lambda: contains("smth"), ContainMatcher),
    (lambda: not_contains("smth"), NotContainMatcher),
    (lambda: regex(r".*"), RegexMatcher),
])
def test_is_instance_of(instance_factory, instance_class):
    with given:
        matcher = instance_factory()

    with when:
        actual = isinstance(matcher, instance_class)

    with then:
        assert actual is True


@pytest.mark.parametrize(("instance_factory", "representation"), [
    (lambda: exists, "exists"),
    (lambda: equals("smth"), "equals('smth')"),
    (lambda: not_equals("smth"), "not_equals('smth')"),
    (lambda: contains("smth"), "contains('smth')"),
    (lambda: not_contains("smth"), "not_contains('smth')"),
    (lambda: regex(r".*"), "regex('.*')"),
])
def test_repr(instance_factory, representation):
    with given:
        matcher = instance_factory()

    with when:
        actual = repr(matcher)

    with then:
        assert actual == representation


@pytest.mark.parametrize("instance_factory", [
    lambda: exists,
    lambda: equals("smth"),
    lambda: not_equals("smth"),
    lambda: contains("smth"),
    lambda: not_contains("smth"),
    lambda: regex(r".*"),
])
def test_packable(instance_factory):
    with given:
        matcher = instance_factory()

    with when:
        actual = pack(matcher)

    with then:
        assert isinstance(actual, bytes)



import logging


class TestLogRecord(logging.LogRecord):
    def __init__(self, message):
        super().__init__(
            name=None,
            level=None,
            pathname=None,
            lineno=None,
            msg=message,
            args=None,
            exc_info=None)



from ._matcher_function import MatcherFunction
from ._registry import Registry
from ._resolver import Resolver
from ._reversed_resolver import ReversedResolver

__all__ = ("MatcherFunction", "Registry", "Resolver", "ReversedResolver",)



__all__ = ("given", "when", "then",)


class Context:
    def __enter__(self):
        pass

    def __exit__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return self


class Given(Context):
    pass


class When(Context):
    pass


class Then(Context):
    pass


given = Given()
when = When()
then = Then()



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

from unittest.mock import Mock, call, sentinel

import pytest

import jj
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher, PathMatcher
from jj.middlewares import AbstractMiddleware, BaseMiddleware, RootMiddleware
from jj.requests import Request
from jj.resolvers import Registry, ReversedResolver
from jj.responses import Response

from .._test_utils import run


class TestBaseMiddleware(TestCase):
    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    def test_middleware_without_resolver(self):
        with self.assertRaises(Exception):
            BaseMiddleware()

    def test_middleware_with_instance_resolver(self):
        middleware = BaseMiddleware(self.resolver)
        self.assertIsInstance(middleware, AbstractMiddleware)

    def test_middleware_with_class_resolver(self):
        class Middleware(BaseMiddleware):
            resolver = self.resolver
        middleware = Middleware()
        self.assertIsInstance(middleware, AbstractMiddleware)

    @pytest.mark.asyncio
    async def test_middleware_without_impl_and_without_handlers(self):
        @BaseMiddleware(self.resolver)
        class App(jj.App):
            resolver = self.resolver

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 404)

    @pytest.mark.asyncio
    async def test_middleware_without_impl_but_with_handlers(self):
        @BaseMiddleware(self.resolver)
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(status=200)

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

    # App middleware

    @pytest.mark.asyncio
    async def test_app_middleware_without_handlers(self):
        mock = Mock()

        class Middleware(BaseMiddleware):
            async def do(self, request, handler, app):
                mock(request, handler, app)
                return await handler(request)

        @Middleware(self.resolver)
        class App(jj.App):
            resolver = self.resolver

        app = App()
        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 404)

        mock.assert_called_once()
        request_arg, handler_arg, app_arg = mock.call_args[0]
        self.assertIsInstance(request_arg, Request)
        self.assertEqual(handler_arg, default_handler)
        self.assertEqual(app_arg, app)

    @pytest.mark.asyncio
    async def test_app_middleware_with_handlers(self):
        mock = Mock()

        class Middleware(BaseMiddleware):
            async def do(self, r, h, a):
                mock(r, h, a)
                return await h(r)

        @Middleware(self.resolver)
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(status=200)

        app = App()
        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_called_once()
        request_arg, handler_arg, app_arg = mock.call_args[0]
        self.assertIsInstance(request_arg, Request)
        self.assertEqual(handler_arg, App.handler)
        self.assertEqual(app_arg, app)

    @pytest.mark.asyncio
    async def test_multiple_app_middlewares(self):
        mock = Mock()

        class Middleware1(BaseMiddleware):
            async def do(self, request, handler, app):
                mock(self.__class__.__name__, sentinel.BEFORE)
                response = await handler(request)
                mock(self.__class__.__name__, sentinel.AFTER)
                return response

        class Middleware2(Middleware1):
            pass

        @Middleware1(self.resolver)
        @Middleware2(self.resolver)
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                mock(App.__name__, sentinel.BEFORE)
                response = Response(status=200)
                mock(App.__name__, sentinel.AFTER)
                return response

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_has_calls([
            call(Middleware2.__name__, sentinel.BEFORE),
            call(Middleware1.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.AFTER),
            call(Middleware1.__name__, sentinel.AFTER),
            call(Middleware2.__name__, sentinel.AFTER),
        ])
        self.assertEqual(mock.call_count, 6)

    @pytest.mark.asyncio
    async def test_app_middleware_hook(self):
        mock = Mock()

        class Middleware(BaseMiddleware):
            def on_app(self, app):
                mock(app)

            def on_handler(self, handler):
                mock(handler)

        @Middleware(self.resolver)
        class App(jj.App):
            resolver = self.resolver

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 404)

        mock.assert_called_once_with(App)

    # Handler middleware

    @pytest.mark.asyncio
    async def test_handler_middleware(self):
        mock = Mock()

        class Middleware(BaseMiddleware):
            async def do(self, r, h, a):
                mock(r, h, a)
                return await h(r)

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher("/path", resolver=resolver)
            @Middleware(resolver)
            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(status=200)

        app = App()
        async with run(app) as client:
            response = await client.get("/path")
            self.assertEqual(response.status, 200)

        mock.assert_called_once()
        request_arg, handler_arg, app_arg = mock.call_args[0]
        self.assertIsInstance(request_arg, Request)
        self.assertEqual(handler_arg, App.handler)
        self.assertEqual(app_arg, app)

    @pytest.mark.asyncio
    async def test_handler_middleware_hook(self):
        mock = Mock()

        class Middleware(BaseMiddleware):
            def on_app(self, app):
                mock(app)

            def on_handler(self, handler):
                mock(handler)

        class App(jj.App):
            resolver = self.resolver

            @Middleware(resolver)
            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(status=200)

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_called_once_with(App.handler)

    @pytest.mark.asyncio
    async def test_multiple_handler_middlewares(self):
        mock = Mock()

        class Middleware1(BaseMiddleware):
            async def do(self, request, handler, app):
                mock(self.__class__.__name__, sentinel.BEFORE)
                response = await handler(request)
                mock(self.__class__.__name__, sentinel.AFTER)
                return response

        class Middleware2(Middleware1):
            pass

        class App(jj.App):
            resolver = self.resolver

            @Middleware1(resolver)
            @MethodMatcher("*", resolver=resolver)
            @Middleware2(resolver)
            async def handler(request):
                mock(App.__name__, sentinel.BEFORE)
                response = Response(status=200)
                mock(App.__name__, sentinel.AFTER)
                return response

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_has_calls([
            call(Middleware2.__name__, sentinel.BEFORE),
            call(Middleware1.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.AFTER),
            call(Middleware1.__name__, sentinel.AFTER),
            call(Middleware2.__name__, sentinel.AFTER),
        ])
        self.assertEqual(mock.call_count, 6)

    # App + Handler middlewares

    @pytest.mark.asyncio
    async def test_app_and_handler_middlewares_priority(self):
        mock = Mock()

        class Middleware1(BaseMiddleware):
            async def do(self, request, handler, app):
                mock(self.__class__.__name__, sentinel.BEFORE)
                response = await handler(request)
                mock(self.__class__.__name__, sentinel.AFTER)
                return response

        class Middleware2(Middleware1):
            pass

        @Middleware1(self.resolver)
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            @Middleware2(resolver)
            async def handler(request):
                mock(App.__name__, sentinel.BEFORE)
                response = Response(status=200)
                mock(App.__name__, sentinel.AFTER)
                return response

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_has_calls([
            call(Middleware1.__name__, sentinel.BEFORE),
            call(Middleware2.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.AFTER),
            call(Middleware2.__name__, sentinel.AFTER),
            call(Middleware1.__name__, sentinel.AFTER),
        ])
        self.assertEqual(mock.call_count, 6)

    # Root + App + Handler middlewares

    @pytest.mark.asyncio
    async def test_root_and_app_and_handler_middlewares_priority(self):
        mock = Mock()

        class Do:
            async def do(self, request, handler, app):
                mock(self.__class__.__name__, sentinel.BEFORE)
                response = await handler(request)
                mock(self.__class__.__name__, sentinel.AFTER)
                return response

        class Middleware(Do, RootMiddleware):
            resolver = self.resolver

        class AppMiddleware(Do, BaseMiddleware):
            resolver = self.resolver

        class HandlerMiddleware(Do, BaseMiddleware):
            resolver = self.resolver

        @AppMiddleware()
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            @HandlerMiddleware()
            async def handler(request):
                mock(App.__name__, sentinel.BEFORE)
                response = Response(status=200)
                mock(App.__name__, sentinel.AFTER)
                return response

        async with run(App(), middlewares=[Middleware()]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_has_calls([
            call(AppMiddleware.__name__, sentinel.BEFORE),
            call(HandlerMiddleware.__name__, sentinel.BEFORE),
            call(Middleware.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.BEFORE),
            call(App.__name__, sentinel.AFTER),
            call(Middleware.__name__, sentinel.AFTER),
            call(HandlerMiddleware.__name__, sentinel.AFTER),
            call(AppMiddleware.__name__, sentinel.AFTER),
        ])
        self.assertEqual(mock.call_count, 8)



import pytest
from aiohttp import FormData

import jj
from jj.middlewares import SelfMiddleware
from jj.mock import Mock, RemoteMock

from .._test_utils import run


@pytest.mark.asyncio
async def test_mock_history_request():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp = await client.get("/", params={"key": "val"})
            assert resp.status == 200

            history = await handler.fetch_history()
            req = history[0]["request"]
            assert req.method == "GET"
            assert req.segments == {}
            assert req.path == "/"
            assert req.params == {"key": "val"}
            assert req.headers.get("User-Agent")
            assert req.body == req.raw == b""

            res = history[0]["response"]
            assert res.status == 200
            assert res.headers.get("Server")
            assert res.reason == "OK"
            assert res.body == res.raw == b"text"


@pytest.mark.asyncio
async def test_mock_history_multiple_requests():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response()

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp1 = await client.get("/path1")
            assert resp1.status == 200

            resp2 = await client.get("/path2")
            assert resp2.status == 200

            history = await handler.fetch_history()
            req1 = history[-1]["request"]
            assert req1.method == "GET"
            assert req1.path == "/path1"

            req = history[0]["request"]
            assert req.method == "GET"
            assert req.path == "/path2"


@pytest.mark.asyncio
async def test_mock_history_post_no_data():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response()

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp = await client.post("/")
            assert resp.status == 200

            history = await handler.fetch_history()
            req = history[0]["request"]
            assert req.method == "POST"
            assert req.path == "/"
            assert req.body == req.raw == b""


@pytest.mark.asyncio
async def test_mock_history_post_json():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response()

    payload = {
        "field1": "value1",
        "field2": None,
    }
    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp = await client.post("/", json=payload)
            assert resp.status == 200

            history = await handler.fetch_history()
            req = history[0]["request"]
            assert req.method == "POST"
            assert req.path == "/"
            assert req.body == payload
            assert req.raw == b'{"field1": "value1", "field2": null}'


@pytest.mark.asyncio
async def test_mock_history_post_invalid_json():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response()

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp = await client.post("/", data="[010]",
                                     headers={"Content-Type": "application/json"})
            assert resp.status == 200

            history = await handler.fetch_history()
            req = history[0]["request"]
            assert req.method == "POST"
            assert req.path == "/"
            assert req.body == req.raw == b"[010]"


@pytest.mark.asyncio
async def test_mock_history_post_data():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response()

    payload = FormData([
        ("field1", "value1"),
        ("field1", "value2"),
        ("field2", "null"),
    ])
    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp = await client.post("/", data=payload)
            assert resp.status == 200

            history = await handler.fetch_history()
            req = history[0]["request"]
            assert req.method == "POST"
            assert req.path == "/"
            assert req.body == req.raw == b'field1=value1&field1=value2&field2=null'


@pytest.mark.asyncio
async def test_mock_history_post_form_data():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response()

    payload = FormData([
        ("field1", "value1"),
        ("field1", "value2"),
        ("field2", "null"),
    ])
    payload._is_multipart = True

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp = await client.post("/", data=payload)
            assert resp.status == 200

            history = await handler.fetch_history()
            req = history[0]["request"]
            assert req.method == "POST"
            assert req.path == "/"
            assert isinstance(req.body, bytes)
            assert isinstance(req.raw, bytes)


@pytest.mark.asyncio
async def test_mock_history_post_binary_data():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response()

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp = await client.post("/", data=b"binary")
            assert resp.status == 200

            history = await handler.fetch_history()
            req = history[0]["request"]
            assert req.method == "POST"
            assert req.path == "/"
            assert req.body == req.raw == b"binary"


@pytest.mark.asyncio
async def test_mock_history_post_text_data():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response()

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(matcher, response)
        async with handler:
            resp = await client.post("/", data="text")
            assert resp.status == 200

            history = await handler.fetch_history()
            req = history[0]["request"]
            assert req.method == "POST"
            assert req.path == "/"
            assert req.body == "text"
            assert req.raw == b"text"



from ..requests import Request

__all__ = ("ExpirationPolicy",)


class ExpirationPolicy:
    async def is_expired(self, request: Request) -> bool:
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"



from types import TracebackType
from typing import TYPE_CHECKING, Any, List, Optional, Type, Union
from uuid import UUID, uuid4

from jj.expiration_policy import ExpirationPolicy
from jj.matchers import LogicalMatcher, RequestMatcher

if TYPE_CHECKING:
    from ._remote_mock import RemoteMock
    MockType = RemoteMock
else:
    MockType = Any

from ._history import HistoryAdapterType, HistoryItem, default_history_adapter
from ._remote_response import RemoteResponseType

__all__ = ("RemoteHandler",)


class RemoteHandler:
    def __init__(self,
                 mock: MockType,
                 matcher: Union[RequestMatcher, LogicalMatcher],
                 response: RemoteResponseType,
                 expiration_policy: Optional[ExpirationPolicy] = None,
                 *,
                 history_adapter: Optional[HistoryAdapterType] = default_history_adapter) -> None:
        self._id = uuid4()
        self._mock = mock
        self._matcher = matcher
        self._response = response
        self._history_adapter = history_adapter
        self._expiration_policy = expiration_policy

    @property
    def id(self) -> UUID:
        return self._id

    @property
    def matcher(self) -> Union[RequestMatcher, LogicalMatcher]:
        return self._matcher

    @property
    def response(self) -> RemoteResponseType:
        return self._response

    @property
    def expiration_policy(self) -> Optional[ExpirationPolicy]:
        return self._expiration_policy

    async def register(self) -> None:
        await self._mock.register(self)

    async def deregister(self) -> None:
        await self._mock.deregister(self)

    async def fetch_history(self) -> List[HistoryItem]:
        history = await self._mock.fetch_history(self)
        if self._history_adapter:
            return [self._history_adapter(x) for x in history]
        return history

    async def __aenter__(self) -> None:
        await self.register()

    async def __aexit__(self,
                        exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> None:
        await self.deregister()



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call, sentinel

import pytest

from jj.matchers import AttributeMatcher, PathMatcher, RequestMatcher
from jj.matchers.attribute_matchers import RouteMatcher

from ..._test_utils.fixtures import request_, resolver_
from ..._test_utils.steps import given, then, when

__all__ = ("request_", "resolver_",)


@pytest.mark.asyncio
@pytest.mark.parametrize(("expected", "actual", "res"), [
    ("/", "/", True),
    ("/", "/smth", False),
    ("/smth", "/", False),

    ("/users/{id}", "/users/1", True),
    ("/users/{id}", "/users", False),
    ("/users/{id}", "/users/", False),
    ("/users/{id}", "/users/1/profile", False),

    ("/{tail:.*}", "/", True),
    ("/{tail:.*}", "/users", True),
    ("/{tail:.*}", "/users/1/profile", True),
])
async def test_path_matcher(expected, actual, res, *, resolver_, request_):
    with given:
        request_.path = actual
        matcher = PathMatcher(expected, resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is res


@pytest.mark.asyncio
@pytest.mark.parametrize(("ret_val", "path"), [
    (True, "/"),
    (True, "/smth"),
    (False, "/"),
])
async def test_path_matcher_with_custom_submatcher(ret_val, path, *, resolver_, request_):
    with given:
        request_.path = path
        submatcher_ = Mock(AttributeMatcher, match=AsyncMock(return_value=ret_val))
        matcher = PathMatcher(submatcher_, resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is ret_val
        assert submatcher_.mock_calls == [call.match(path)]


def test_is_instance_of_request_matcher(*, resolver_):
    with given:
        matcher = PathMatcher("/", resolver=resolver_)

    with when:
        actual = isinstance(matcher, RequestMatcher)

    with then:
        assert actual is True


@pytest.mark.parametrize("path", ["/", "/users", "/users/{id}"])
def test_repr(path, *, resolver_):
    with given:
        resolver_.__repr__ = Mock(return_value="<Resolver>")
        matcher = PathMatcher(path, resolver=resolver_)

    with when:
        actual = repr(matcher)

    with then:
        assert actual == f"PathMatcher(RouteMatcher({path!r}), resolver=<Resolver>)"


def test_pack():
    with given:
        submatcher = RouteMatcher("/")
        matcher = PathMatcher(submatcher, resolver=resolver_)

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {"path": submatcher}


def test_unpack():
    with given:
        submatcher = RouteMatcher("/")
        kwargs = {
            "path": submatcher,
            "resolver": resolver_,
            "future_field": sentinel,
        }

    with when:
        actual = PathMatcher.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, PathMatcher)



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
    from unittest.mock import AsyncMock
else:
    from unittest import TestCase
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call, sentinel

import pytest

from jj.apps import create_app
from jj.resolvers import Registry, ReversedResolver


class TestReversedResolver(TestCase):
    def setUp(self):
        self.default_handler = AsyncMock(return_value=sentinel.default_response)
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, self.default_handler)

    def test_handler_getter_without_handlers(self):
        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [])

    def test_handler_getter_with_one_handler(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))

        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [handler])

    def test_handler_getter_with_multiple_handlers(self):
        handler1 = AsyncMock(return_value=sentinel.response)
        handler2 = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler1, type(self.default_app))
        self.resolver.register_handler(handler2, type(self.default_app))

        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [handler2, handler1])

    @pytest.mark.asyncio
    async def test_resolve_request_with_multiple_handlers(self):
        matcher = AsyncMock(side_effect=(False, True))
        handler1 = AsyncMock(return_value=sentinel.response1)
        handler2 = AsyncMock(return_value=sentinel.response2)
        self.resolver.register_matcher(matcher, handler1)
        self.resolver.register_matcher(matcher, handler2)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, handler2)

        handler1.assert_not_called()
        handler2.assert_not_called()
        matcher.assert_has_calls([call(request)] * 2, any_order=True)
        self.assertEqual(matcher.call_count, 2)

    @pytest.mark.asyncio
    async def test_resolve_request_priority(self):
        matcher = AsyncMock(side_effect=(True, True))
        handler1 = AsyncMock(return_value=sentinel.response1)
        handler2 = AsyncMock(return_value=sentinel.response2)
        self.resolver.register_matcher(matcher, handler1)
        self.resolver.register_matcher(matcher, handler2)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, handler1)

        handler1.assert_not_called()
        handler2.assert_not_called()
        matcher.assert_called_once_with(request)



from typing import Any, Dict

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from ..attribute_matchers import AttributeMatcher, EqualMatcher, StrOrAttrMatcher
from ._request_matcher import RequestMatcher

__all__ = ("MethodMatcher",)


@packable("jj.matchers.MethodMatcher")
class MethodMatcher(RequestMatcher):
    def __init__(self, method: StrOrAttrMatcher, *, resolver: Resolver) -> None:
        super().__init__(resolver=resolver)
        if isinstance(method, AttributeMatcher):
            self._matcher = method
        else:
            self._matcher = EqualMatcher(str.upper(method))

    async def match(self, request: Request) -> bool:
        return await self._matcher.match("*") or await self._matcher.match(request.method)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._matcher!r}, resolver={self._resolver!r})"

    def __packed__(self) -> Dict[str, Any]:
        return {"method": self._matcher}

    @classmethod
    def __unpacked__(cls, *,
                     method: StrOrAttrMatcher,
                     resolver: Resolver,
                     **kwargs: Any) -> "MethodMatcher":
        return cls(method, resolver=resolver)



from unittest.mock import Mock, call, sentinel

import pytest

from jj.logs import Filter

from .._test_utils.steps import given, then, when
from ._log_record import TestLogRecord


@pytest.fixture()
def record():
    return TestLogRecord(sentinel.message)


def test_filter_without_request_and_response(record: TestLogRecord):
    with given:
        filter = Filter()
        filter.filter_request = Mock(return_value=False)
        filter.filter_response = Mock(return_value=False)

    with when:
        res = filter.filter(record)

    with then:
        assert res is True
        assert filter.filter_request.mock_calls == []
        assert filter.filter_response.mock_calls == []


@pytest.mark.parametrize("value", [False, True])
def test_filter_with_request(value: bool, record: TestLogRecord):
    with given:
        filter = Filter()
        filter.filter_request = Mock(return_value=value)
        filter.filter_response = Mock(return_value=True)
        record.jj_request = sentinel.request

    with when:
        res = filter.filter(record)

    with then:
        assert res is value
        assert filter.filter_request.mock_calls == [
            call(sentinel.request, record)
        ]
        assert filter.filter_response.mock_calls == []


@pytest.mark.parametrize("value", [False, True])
def test_filter_with_response(value: bool, record: TestLogRecord):
    with given:
        filter = Filter()
        filter.filter_request = Mock(return_value=True)
        filter.filter_response = Mock(return_value=value)
        record.jj_request = sentinel.request
        record.jj_response = sentinel.response

    with when:
        res = filter.filter(record)

    with then:
        assert res is value
        assert filter.filter_request.mock_calls == []
        assert filter.filter_response.mock_calls == [
            call(sentinel.response, sentinel.request, record)
        ]


def test_request_filter(record: TestLogRecord):
    with given:
        filter = Filter()

    with when:
        res = filter.filter_request(sentinel.request, record)

    with then:
        assert res is True


def test_response_filter(record: TestLogRecord):
    with given:
        filter = Filter()

    with when:
        res = filter.filter_response(sentinel.response, sentinel.request, record)

    with then:
        assert res is True



from typing import Any, Dict

from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("EqualMatcher", "NotEqualMatcher",)


@packable("jj.matchers.EqualMatcher")
class EqualMatcher(AttributeMatcher):
    def __init__(self, expected: Any) -> None:
        self._expected = expected

    async def match(self, actual: Any) -> bool:
        return bool(self._expected == actual)  # type casting for mypy

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._expected!r})"

    def __packed__(self) -> Dict[str, Any]:
        return {"expected": self._expected}

    @classmethod
    def __unpacked__(cls, *, expected: Any, **kwargs: Any) -> "EqualMatcher":
        return cls(expected)


@packable("jj.matchers.NotEqualMatcher")
class NotEqualMatcher(EqualMatcher):
    async def match(self, actual: Any) -> bool:
        return bool(self._expected != actual)  # type casting for mypy

    def __packed__(self) -> Dict[str, Any]:
        return {"expected": self._expected}

    @classmethod
    def __unpacked__(cls, *, expected: Any, **kwargs: Any) -> "NotEqualMatcher":
        return cls(expected)



from unittest.mock import Mock, sentinel

import pytest

from jj.logs import SimpleFormatter

from .._test_utils.steps import given, then, when
from ._log_record import TestLogRecord


@pytest.fixture()
def formatter():
    return SimpleFormatter()


@pytest.fixture()
def record():
    return TestLogRecord(sentinel.message)


def test_format_without_request_and_response(formatter: SimpleFormatter, record: TestLogRecord):
    with when:
        res = formatter.format(record)

    with then:
        assert res == str(sentinel.message)


def test_format_with_request(formatter: SimpleFormatter, record: TestLogRecord):
    with given:
        record.jj_request = Mock(method=sentinel.method, url=Mock(path=sentinel.path))

    with when:
        res = formatter.format(record)

    with then:
        assert res == "-> {} {}".format(sentinel.method, sentinel.path)


def test_format_with_response(formatter: SimpleFormatter, record: TestLogRecord):
    with given:
        record.jj_request = Mock()
        record.jj_response = Mock(status=sentinel.status, reason=sentinel.reason)

    with when:
        res = formatter.format(record)

    with then:
        assert res == "<- {} {}\n".format(sentinel.status, sentinel.reason)



from unittest.mock import Mock, sentinel

import pytest

from jj.mock import SystemLogFilter

from .._test_utils.steps import given, then, when
from ..logs._log_record import TestLogRecord


@pytest.fixture()
def record():
    return TestLogRecord(sentinel.message)


def test_log_filter(record: TestLogRecord):
    with given:
        log_filter = SystemLogFilter()

    with when:
        res = log_filter.filter(record)

    with then:
        assert res is True


def test_log_filter_request_without_header(record: TestLogRecord):
    with given:
        log_filter = SystemLogFilter()
        record.jj_request = Mock(headers={})

    with when:
        res = log_filter.filter(record)

    with then:
        assert res is True


def test_log_filter_request_with_header(record: TestLogRecord):
    with given:
        log_filter = SystemLogFilter()
        record.jj_request = Mock(headers={"x-jj-remote-mock": ""})

    with when:
        res = log_filter.filter(record)

    with then:
        assert res is False


def test_log_filter_response_without_header(record: TestLogRecord):
    with given:
        log_filter = SystemLogFilter()
        record.jj_request = Mock(headers={})
        record.jj_response = Mock()

    with when:
        res = log_filter.filter(record)

    with then:
        assert res is True


def test_log_filter_response_with_header(record: TestLogRecord):
    with given:
        log_filter = SystemLogFilter()
        record.jj_request = Mock(headers={"x-jj-remote-mock": ""})
        record.jj_response = Mock()

    with when:
        res = log_filter.filter(record)

    with then:
        assert res is False



from typing import Any, Dict, List

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from .._resolvable_matcher import ResolvableMatcher
from ._logical_matcher import LogicalMatcher

__all__ = ("AnyMatcher",)


@packable("jj.matchers.AnyMatcher")
class AnyMatcher(LogicalMatcher):
    def __init__(self, matchers: List[ResolvableMatcher], *, resolver: Resolver) -> None:
        super().__init__(resolver=resolver)
        assert len(matchers) > 0
        self._matchers = matchers

    async def match(self, request: Request) -> bool:
        for matcher in self._matchers:
            if await matcher.match(request):
                return True
        return False

    def __repr__(self) -> str:
        return (f"{self.__class__.__qualname__}"
                f"({self._matchers!r}, resolver={self._resolver!r})")

    def __packed__(self) -> Dict[str, Any]:
        return {"matchers": self._matchers}

    @classmethod
    def __unpacked__(cls, *,
                     matchers: List[ResolvableMatcher],
                     resolver: Resolver,
                     **kwargs: Any) -> "AnyMatcher":
        return cls(matchers, resolver=resolver)



import asyncio
from functools import partial
from typing import List, Optional, Sequence, Type

from aiohttp import web

from .apps import AbstractApp, BaseApp, DefaultApp
from .handlers import default_handler
from .logs import Logger
from .logs import logger as default_logger
from .matchers import ResolvableMatcher
from .matchers.logical_matchers import AllMatcher, AnyMatcher
from .matchers.request_matchers import (
    DictOrTupleListOrAttrMatcher,
    HeaderMatcher,
    MethodMatcher,
    ParamMatcher,
    PathMatcher,
    StrOrAttrMatcher,
)
from .middlewares import BaseMiddleware, LoggerMiddleware, SelfMiddleware
from .resolvers import Registry, ReversedResolver
from .runners import AppRunner
from .servers import Server

__all__ = (
    "App",
    "Middleware",
    "match_method", "match_methods",
    "match_path",
    "match_headers", "match_header",
    "match_params", "match_param",
    "match_all", "match_any",
    "match",
    "start", "wait_for", "serve",
    "default_app", "default_handler", "default_logger",
)

default_app = DefaultApp()
registry = Registry()
resolver = ReversedResolver(registry, default_app, default_handler)
loop = asyncio.get_event_loop()

runner = partial(AppRunner, resolver=resolver, middlewares=[
    SelfMiddleware(resolver),
])
# ignore because of https://github.com/python/mypy/issues/1484
server = Server(loop, runner, web.TCPSite)  # type: ignore


class App(BaseApp):
    resolver = resolver


class Middleware(BaseMiddleware):
    resolver = resolver


def match_method(method: StrOrAttrMatcher) -> MethodMatcher:
    return MethodMatcher(method, resolver=resolver)


def match_methods(*methods: StrOrAttrMatcher) -> AnyMatcher:
    matchers: List[ResolvableMatcher] = [
        MethodMatcher(method, resolver=resolver)
        for method in methods
    ]
    return AnyMatcher(matchers, resolver=resolver)


def match_path(path: StrOrAttrMatcher) -> PathMatcher:
    return PathMatcher(path, resolver=resolver)


def match_headers(headers: DictOrTupleListOrAttrMatcher) -> HeaderMatcher:
    return HeaderMatcher(headers, resolver=resolver)


def match_header(name: str, value: StrOrAttrMatcher) -> HeaderMatcher:
    return HeaderMatcher([(name, value)], resolver=resolver)


def match_params(params: DictOrTupleListOrAttrMatcher) -> ParamMatcher:
    return ParamMatcher(params, resolver=resolver)


def match_param(name: str, value: StrOrAttrMatcher) -> ParamMatcher:
    return ParamMatcher([(name, value)], resolver=resolver)


def match_all(matchers: List[ResolvableMatcher]) -> AllMatcher:
    return AllMatcher(matchers, resolver=resolver)


def match_any(matchers: List[ResolvableMatcher]) -> AnyMatcher:
    return AnyMatcher(matchers, resolver=resolver)


def match(method: Optional[StrOrAttrMatcher] = None,
          path: Optional[StrOrAttrMatcher] = None,
          params: Optional[DictOrTupleListOrAttrMatcher] = None,
          headers: Optional[DictOrTupleListOrAttrMatcher] = None) -> AllMatcher:
    submatchers: List[ResolvableMatcher] = []
    if method:
        submatchers += [MethodMatcher(method, resolver=resolver)]
    if path:
        submatchers += [PathMatcher(path, resolver=resolver)]
    if params:
        submatchers += [ParamMatcher(params, resolver=resolver)]
    if headers:
        submatchers += [HeaderMatcher(headers, resolver=resolver)]
    return AllMatcher(submatchers, resolver=resolver)


def start(app: AbstractApp, *,
          host: Optional[str] = None, port: Optional[int] = None,
          logger: Logger = default_logger) -> None:
    LoggerMiddleware(resolver, logger)(type(app))
    server.start(app, host=host, port=port)


def wait_for(exceptions: Sequence[Type[BaseException]]) -> None:
    try:
        server.serve()
    except tuple(exceptions):
        pass
    finally:
        server.cleanup()
    server.shutdown()


def serve(app: AbstractApp = default_app, *,
          host: Optional[str] = None, port: Optional[int] = None,
          logger: Logger = default_logger) -> None:
    start(app, host=host, port=port, logger=logger)
    wait_for([KeyboardInterrupt])
